@@ -15,7 +15,7 @@ __all__ = [
     "impacts", "nllscan",
 ]
 
-def cms_label(ax, label, lumi=35.9, energy=13):
+def cms_label(ax, label, lumi=35.9, energy=13, extra_label=""):
     ax.text(
         0, 1, r'$\mathbf{CMS}\ \mathit{'+label+'}$',
         ha='left', va='bottom', transform=ax.transAxes,
@@ -24,6 +24,11 @@ def cms_label(ax, label, lumi=35.9, energy=13):
         1, 1, r'${:.1f}\ \mathrm{{fb}}^{{-1}}$ ({:.0f} TeV)'.format(lumi, energy),
         ha='right', va='bottom', transform=ax.transAxes,
     )
+    # label on centre top of axes
+    ax.text(
+        0.5, 1, extra_label,
+        ha='center', va='bottom', transform=ax.transAxes,
+    )
 
 def legend_data_mc(
     ax, df_data, df_mc, label, add_ratios=True, offaxis=True, legend_kw={},
@@ -90,15 +95,31 @@ def data(ax, df, label, bins, data_kw={}):
     kwargs = dict(fmt='o', lw=1, color='black', label='Data')
     kwargs.update(data_kw)
 
-    #down, up = poisson_interval(df["count"], scale=df["sum_w"]/df["count"])
+    mask = (df["sum_ww"]==0.)
     neff = df["sum_w"]**2 / df["sum_ww"]
-    down, up = poisson_interval(neff, scale=df["sum_w"]/neff)
+    neff[mask] = 0.
+
+    scale = df["sum_w"]/neff
+    scale[mask] = 1.
+
+    down, up = poisson_interval(neff, scale=scale)
     ax.errorbar(
         bin_cents, df["sum_w"], yerr=[df["sum_w"]-down, up-df["sum_w"]],
         **kwargs,
     )
 
-def mc(ax, df, label, bins, mcstat=False, mc_kw={}, mcstat_kw={}, proc_kw={}):
+def poisson_interval_with_checks(x, variance):
+    down, up = poisson_interval(x**2/variance, scale=variance/x)
+    mask = (variance==0.)
+    down[mask] = 0.
+    up[mask] = np.inf
+    return down, up
+
+def mc(
+    ax, df, label, bins, mcstat=False, mc_kw={}, mcstat_kw={}, proc_kw={},
+    zorder=0, interval_func=poisson_interval_with_checks, sort_by_process=True,
+):
+    stacked = mc_kw.pop("stacked") if "stacked" in mc_kw else False
     bin_edges, bin_cents = bin_lows_to_edges_cents(bins)
 
     # preprocess mc
@@ -108,44 +129,82 @@ def mc(ax, df, label, bins, mcstat=False, mc_kw={}, mcstat_kw={}, proc_kw={}):
     )
 
     # sort by process total
-    tdf_procsum = tdf.sum(axis=0)
-    tdf = tdf[tdf_procsum.sort_values().index]
+    if sort_by_process:
+        tdf_procsum = tdf.sum(axis=0)
+        tdf = tdf[tdf_procsum.sort_values().index]
 
     # mc
     procs = tdf.columns.to_series()
-    kwargs = dict(
-        label=procs.replace(proc_kw.get("labels", {})),
-        color=procs.apply(lambda x: proc_kw.get("colours", {}).get(x, "blue")),
-    )
-    kwargs.update(mc_kw)
-    ax.hist([bin_cents]*tdf.shape[1], bins=bin_edges, weights=tdf.T, **kwargs)
+
+    cumsum = tdf.iloc[:,0].copy(deep=True)
+    cumsum.values[:] = 0.
+    for idx, proc in enumerate(tdf.columns):
+        if stacked:
+            prev_cumsum = cumsum.copy(deep=True)
+            cumsum += tdf[proc]
+        else:
+            cumsum = tdf[proc]
+
+        color = proc_kw.get("colours", {}).get(proc, "blue")
+        kwargs = {
+            "color": color, "ec": color,
+            "label": proc_kw.get("labels", {}).get(proc, proc),
+        }
+        kwargs.update(mc_kw)
+        if stacked:
+            kwargs.update({"ec": color, "lw": 0.1, "zorder": -idx})
+        ax.hist(bin_cents, bins=bin_edges, weights=cumsum, **kwargs)
 
     if mcstat:
-        tdf_ww = pd.pivot_table(
+        tdf_ww_up = pd.pivot_table(
             df, index=label, columns="parent",
-            values="sum_ww", aggfunc=np.sum,
+            values="sum_ww_up", aggfunc=np.sum,
         )
-        neff = tdf**2 / tdf_ww
-        down, up = poisson_interval(neff, scale=tdf/neff)
+        _, up = interval_func(tdf.values[:,0], tdf_ww_up.values[:,0])
+
+        tdf_ww_down = pd.pivot_table(
+            df, index=label, columns="parent",
+            values="sum_ww_down", aggfunc=np.sum,
+        )
+        down, _ = interval_func(tdf.values[:,0], tdf_ww_down.values[:,0])
+
         kwargs = dict(color='black', alpha=0.2)
         kwargs.update(mcstat_kw)
-        up = up[:,0]
-        down = down[:,0]
+
         ax.fill_between(
-            bin_edges, list(up)+[up[-1]], list(down)+[down[-1]],
+            bin_edges, list(up)+[list(up)[-1]],
+            list(down)+[list(down)[-1]],
             step='post', **kwargs
         )
 
 def data_mc(
-    ax, df_data, df_mc, label, bins, blind=False, log=True, legend=True,
-    ratio=True, sm_total=True, mcstat_top=False, add_ratios=True, mc_kw={},
-    mcstat_kw={}, sm_kw={}, data_kw={}, proc_kw={}, legend_kw={}, cms_kw={},
+    ax, df_data, df_mc, label, bins,
+    sigs=[], blind=False, log=True, legend=True, ratio=True, sm_total=True,
+    mcstat_top=False, mcstat=True, add_ratios=True, show_zeros=False,
+    mc_kw={}, sig_kw={}, mcstat_kw={}, sm_kw={}, data_kw={}, proc_kw={},
+    legend_kw={}, cms_kw={}, interval_func=poisson_interval_with_checks,
 ):
-    if blind:
-        df_data = df_data*0.
+    _df_data = df_data.copy(deep=True)
+    _df_mc = df_mc.copy(deep=True)
+
+    if not show_zeros:
+        _df_data.loc[_df_data["sum_w"]==0.,"sum_w"] = np.nan
+
+    # only mc sum_ww can be asymmetric
+    if "sum_ww_up" not in _df_mc:
+        _df_mc["sum_ww_up"] = _df_mc["sum_ww"]
+    if "sum_ww_down" not in _df_mc:
+        _df_mc["sum_ww_down"] = _df_mc["sum_ww"]
+
+    # collect signals if set
+    sigs = sigs[::-1]
+    sig_mask = ~_df_mc.index.get_level_values("parent").isin(sigs)
+    df_sig = _df_mc.loc[~sig_mask].copy(deep=True)
+
+    df_mc_sm = _df_mc.loc[sig_mask].copy(deep=True)
 
     # preprocessing
-    df_mc_sum = df_mc.groupby(label).sum()
+    df_mc_sum = df_mc_sm.groupby(label).sum()
     df_mc_sum.loc[:,"parent"] = "SMTotal"
     df_mc_sum = df_mc_sum.groupby(["parent", label]).sum()
 
@@ -156,12 +215,21 @@ def data_mc(
     bin_edges, _ = bin_lows_to_edges_cents(bins)
     ax[0].set_xlim(bin_edges.min(), bin_edges.max())
 
+    # signals - top panel
+    sig_kw_ = dict(histtype='step', zorder=10)
+    sig_kw_.update(sig_kw)
+    if len(sigs) > 0:
+        mc(
+            ax[0], df_sig, label, bins, mcstat=False, mc_kw=sig_kw_,
+            proc_kw=proc_kw, interval_func=interval_func, sort_by_process=False,
+        )
+
     # MC - top panel
     mc_kw_ = dict(stacked=True)
     mc_kw_.update(mc_kw)
     mc(
-        ax[0], df_mc, label, bins, mcstat=mcstat_top, mc_kw=mc_kw_,
-        proc_kw=proc_kw,
+        ax[0], df_mc_sm, label, bins, mcstat=False,
+        mc_kw=mc_kw_, proc_kw=proc_kw, interval_func=interval_func,
     )
 
     # SM total - top panel
@@ -171,12 +239,13 @@ def data_mc(
         mcstat_kw_ = dict(label="", color="black", alpha=0.2)
         mcstat_kw_.update(mcstat_kw)
         mc(
-            ax[0], df_mc_sum, label, bins, mcstat=False, mc_kw=mc_kw_,
-            mcstat_kw=mcstat_kw_, proc_kw=proc_kw,
+            ax[0], df_mc_sum, label, bins, mcstat=mcstat_top, mc_kw=mc_kw_,
+            mcstat_kw=mcstat_kw_, proc_kw=proc_kw, interval_func=interval_func,
         )
 
     # Data - top panel
-    data(ax[0], df_data, label, bins, data_kw=data_kw)
+    if not blind:
+        data(ax[0], _df_data, label, bins, data_kw=data_kw)
 
     # CMS label - top panel
     kwargs = dict(label="Preliminary", lumi=35.9, energy=13)
@@ -185,28 +254,32 @@ def data_mc(
 
     # SM total ratio - bottom panel
     df_mc_sum_ratio = df_mc_sum.copy()
-    #df_mc_sum_ratio.loc[:,"count"] = df_mc_sum["count"]
     df_mc_sum_ratio.loc[:,"sum_w"] = 1.
-    df_mc_sum_ratio.loc[:,"sum_ww"] = df_mc_sum["sum_ww"]/df_mc_sum["sum_w"]**2
+    df_mc_sum_ratio.loc[:,"sum_ww_up"] = (
+        df_mc_sum["sum_ww_up"]/df_mc_sum["sum_w"]**2
+    )
+    df_mc_sum_ratio.loc[:,"sum_ww_down"] = (
+        df_mc_sum["sum_ww_down"]/df_mc_sum["sum_w"]**2
+    )
 
     if ratio:
         mc_kw_ = dict(label="", histtype='step')
         mc_kw_.update(sm_kw)
         mcstat_kw_ = dict(label="MC stat. unc.", color="black", alpha=0.2)
         mcstat_kw_.update(mcstat_kw)
+
         mc(
-            ax[1], df_mc_sum_ratio, label, bins, mcstat=True, mc_kw=mc_kw_,
-            mcstat_kw=mcstat_kw_, proc_kw=proc_kw,
+            ax[1], df_mc_sum_ratio, label, bins, mcstat=mcstat, mc_kw=mc_kw_,
+            mcstat_kw=mcstat_kw_, proc_kw=proc_kw, interval_func=interval_func,
         )
 
         # Data ratio - bottom panel
         if not blind:
             kwargs = dict(data_kw)
             kwargs["label"] = ""
-            df_data_ratio = df_data.copy()
-            #df_data_ratio.loc[:,"count"] = df_data["count"]
-            df_data_ratio.loc[:,"sum_w"] = df_data["sum_w"]/df_mc_sum["sum_w"].values
-            df_data_ratio.loc[:,"sum_ww"] = df_data["sum_ww"]/df_mc_sum["sum_w"].values**2
+            df_data_ratio = _df_data.copy()
+            df_data_ratio.loc[:,"sum_w"] = _df_data["sum_w"]/df_mc_sum["sum_w"].values
+            df_data_ratio.loc[:,"sum_ww"] = _df_data["sum_ww"]/df_mc_sum["sum_w"].values**2
             data(ax[1], df_data_ratio, label, bins, data_kw=kwargs)
 
         if legend:
@@ -214,7 +287,7 @@ def data_mc(
             kwargs = dict(labelspacing=0.05)
             kwargs.update(legend_kw)
             legend_data_mc(
-                ax, df_data, df_mc, label, add_ratios=add_ratios,
+                ax, _df_data, _df_mc, label, add_ratios=add_ratios,
                 offaxis=offaxis, legend_kw=kwargs,
             )
 
@@ -573,8 +646,8 @@ def nllscan(
 
         kw = dict(lw=1, ls='--', color='gray')
         kw.update(line_kw)
-        ax.axvline(left.root, **kw)
-        ax.axvline(right.root, **kw)
+        ax.plot((left.root, left.root), (0., nsig**2), **kw)
+        ax.plot((right.root, right.root), (0., nsig**2), **kw)
         ax.axhline(nsig**2, **kw)
 
         pos = ax.transData.inverted().transform(
@@ -582,6 +655,6 @@ def nllscan(
         )
         kw = dict(ha='left', va='bottom', color='gray')
         kw.update(text_kw)
-        ax.text(pos[0], nsig**2, f'${nsig}\\sigma$', **kw)
+        #ax.text(pos[0], nsig**2, f'${nsig}\\sigma$', **kw)
 
     return pd.DataFrame(outdata)
